require 'logger'
require 'ext_logger/common'
# require 'rails'

class ExtLogger
  include Common

  DATE_FORMAT_MEMBER_LIST = %w(%Y %y %m %d %wday %H %M %S)
  LOG_NAME = '{log_name}'   # Log name pseudo-code

  # Initialize, next is parameters:
  # 1rst: log path (Default parameter)
  # 2end: option (Influencing log size, file name format, etc.)
  #   :age            -- Extention by Logger(shift_age), split file with time format;
  #   :size           -- Extention by Logger(shift_size), split file with fix size;
  #   :split_format   -- Rename logs according to ruby's time conversion format;
  #   :log_max_num    -- Keep the latest number of log files according to the parameter values, and delete the redundancy;
  #   :debug          -- Debug for puts
  def initialize(*args)
    func_name = "[#{self.class.to_s}.initialize]"
    output_debug "#{func_name} args.size: #{args.size}"

    # Instance variables
    # @debug = false          # Controller debug content output to console
    @logger = nil           # Initialize logger object
    @log_path = ''          # Log path
    @log_folder = ''        # Log folder
    @log_file_name = ''     # Log file name
    @log_split_format = ''  # Split format character in log, default blank.
    @log_max_num = 0        # Retain the latest number of log files.
    @output_flag = true     # Controller normal content output to console

    init(args[0]) if args.size == 1

    if args.size == 2
      opt = args[1].clone
      if opt.is_a?(Hash)
        init(args[0], opt)
        args = [args[0]]

        shift_age = is_blank?(opt[:age]) ? opt[:age] : 0
        args << shift_age

        shift_size = is_blank?(opt[:size]) ? opt[:size] : 0
        args << shift_size
      else
        init(args[0])
      end
    end

    # ExtTest.new()
    @logger = Logger.new(*args)
  end

  # Set output flag
  def set_output_flag(type=true)
    @output_flag = type
  end

  # Output info
  def info(*args)
    check_log_file
    output_normal('INFO', *args)

    @logger.info(*args)
  end

  # Output debug
  def debug(*args)
    check_log_file
    output_normal('DEBUG', *args)

    @logger.debug(*args)
  end

  # Output error
  def error(*args)
    check_log_file
    output_normal('ERROR', *args)

    @logger.error(*args)
  end

  # output warning
  def warn(*args)
    check_log_file
    output_normal('WARN', *args)

    @logger.warn(*args)
  end

  # Output fatal
  def fatal(*args)
    check_log_file
    output_normal('FATAL', *args)

    @logger.fatal(*args)
  end

  # Output exception
  def exception(*args)
    func_name = "[#{get_instance_name}.exception]"

    if args.size == 1
      str = analysis_exception(args[0])
      return if is_blank?(str)
      args = [str]
    end
    output_normal('EXCEPTION', *args)

    _output_type = 'error'
    if args.size == 2
      output_debug "#{func_name} args[1].class: #{args[1].class}"
      if args[1].is_a?(Hash)
        _output_type = args[1][:type] if !args[1][:type].blank?
      end
    end

    # Change logger ouput type
    case _output_type
      when 'info'
        @logger.info(*args)
      when 'debug'
        @logger.debug(*args)
      when 'warn'
        @logger.warn(*args)
      when 'fatal'
        @logger.fatal(*args)
      else
        @logger.error(*args)
    end
  end

  # Output the memory usage
  # Example: record_memory()
  # Example: record_memory('end')
  def record_memory(type='start')
    func_name = "[@logger.record_memory]"
    begin
      type_list = %w(start end)
      @record_memory_error ||= 0
      error_message = 'RECORD MEMORY ERROR:'
      message = ''
      if !type_list.include?(type) # Judging whether the parameters are valid
        self.info error_message + ' params error!' if @record_memory_error == 0
        @record_memory_error = 1
        return
      end

      # Get process memory
      if File.exist?("/proc/#{Process.pid}/status")
        process_status = File.open("/proc/#{Process.pid}/status")

        15.times { process_status.gets }

        if type != type_list[1]
          @rss_before_action = process_status.gets.split[1].to_f/1000
          message = "START MEMORY: #{@rss_before_action}MB "
          #@rss_before_action = 0
        else
          @rss_after_action = process_status.gets.split[1].to_f/1000
          message = "END MEMORY: #{@rss_after_action}MB  CONSUME MEMORY: #{@rss_after_action - @rss_before_action}MB "
          #@rss_after_action = 0
        end

        process_status.close
      else
        error_message += " Non Linux system, file [/proc/#{Process.pid}/status] not exist!"
      end

      # Export internal consumption
      self.info message if !message.nil? && message != ''
      if !message.nil? && message != '' && @record_memory_error == 0
        self.info error_message
        @record_memory_error = 1
      end
    rescue Exception=>e
      self.exception e
    end
  end

  private
  # Private function for initialize
  def init(path, opt={})
    func_name = "[#{get_class_name}:private #{__method__}]"
    set_debug(true) if opt[:debug]

    output_debug "#{func_name} path: #{path}, opt: #{opt}"
    _path = path
    set_path(_path)
    # if path.is_a?(String)
    #   if path.include?(ExtLogger.rails_root)
    #     _path = path.gsub(ExtLogger.rails_root, '')
    #     _path = '/log/' + _path if !_path.include?('log/')
    #     _path = _path.gsub('//', '/')
    #   else
    #     _path = path
    #   end
    #   _path = _path.gsub('./', ::Dir.pwd + '/')
    #   set_path(_path)  # Save original log path
    # end

    # Check the folder and create it.
    mkdir_more(_path)

    # Process parameters
    if !is_blank?(opt[:split_format])
      set_split_format(opt[:split_format])
    end

    if !is_blank?(opt[:log_max_num])
      set_log_max_num(opt[:log_max_num])
    end
    output_debug "#{func_name} split_format: #{get_split_format}, log_max_num: #{get_log_max_num}"

    # Delete redundant old log file
    delete_redundant_file
  end

  # Check: Does the log file exist? If not, create it immediately
  def check_log_file
    func_name = "[#{get_class_name}:private check_log_file]"
    _file_path = get_path
    puts "#{func_name} _file_path ### #{_file_path}"
    # File.new(_file_path, 'w+') if !File.exist?(_file_path)
    @logger = Logger.new(_file_path) if !File.exist?(_file_path)
  end

  # Check split_format contain the format member
  def has_date_format_member
    func_name = "[#{get_class_name}:private has_date_format_member]"

    _has_flag = false
    return _has_flag if is_blank?(@log_split_format) || !@log_split_format.include?('%')

    DATE_FORMAT_MEMBER_LIST.each do |item|
      if @log_split_format.include?(item)
        _has_flag = true
        break
      end
    end

    output_debug "#{func_name} @log_split_format: #{@log_split_format}, _has_flag: #{_has_flag}"
    return _has_flag
  end

  # Split log file, doesn't consider file extensions
  def split_file
    func_name = "[#{get_class_name}:private split_file]"
    if has_date_format_member
      _file_path = get_path
      # _file_ext = ''
      #
      # if _file_path.include?('.')
      #   _file_array = _file_path.split('.')
      #   _file_ext   = _file_array.last
      # end

      if File.exist?(_file_path)
        # Get create datetime of current log file
        _file_date = File.ctime(_file_path)
        output_debug "#{func_name} Current log ctime: ##{_file_date}, mtime: #{File.atime(_file_path)}, atime: #{File.mtime(_file_path)}"

        _now_time = Time.now
        _split_format = get_split_format
        _split_format = _split_format.gsub('%S', '').gsub('%s', '') if !get_debug # Non-debug state, does not support seconds splitting files.
        output_debug "#{func_name} split_format: #{_split_format}"
        output_debug "#{func_name} _now_time: #{_now_time.strftime(_split_format)}"
        if _file_date.strftime(_split_format) != _now_time.strftime(_split_format)
          @logger.close

          _tmp_path = _file_path.clone
          # _tmp_path = _tmp_path.gsub(".#{_file_ext}", '') if !is_blank?(_file_ext)
          _old_path = _tmp_path + '.' + _now_time.strftime(_split_format)
          # _old_path += '.' + _file_ext if !is_blank?(_file_ext)

          File.delete(_old_path) if File.exist?(_old_path)
          File.rename(_file_path, _old_path)

          @logger = Logger.new(get_path)
        end
      end
    end
  end

  # Delete redundant old log file
  def delete_redundant_file
    func_name = "[#{get_class_name}:private delete_redundant_file]"
    output_debug "#{func_name} get_log_max_num: #{get_log_max_num}"
    return if get_log_max_num == 0

    _file_list = []
    Dir.entries(get_folder).each do |_file|
      _file_list << _file if _file.match(Regexp.new("^#{get_file_name}"))
    end

    if _file_list.size > 1
      _file_list = _file_list.sort{|a, b| File.ctime("#{get_folder}/#{b}") <=> File.ctime("#{get_folder}/#{a}")}
    end
    output_debug "#{func_name} _file_list: #{_file_list}"

    if _file_list.size >= 1
      while _file_list.size > get_log_max_num
        File.delete("#{get_folder}/#{_file_list.at(-1)}")
        _file_list.delete_at(-1)
        output_debug "#{func_name} _file_list: #{_file_list}"
      end
    end
  end

  # Check output flag & env
  def output_normal(*args)
    func_name = "[#{get_class_name}:private output_normal]"
    _flag = false
    _flag = ExtLogger.rails_env != ENV_HASH[:pro] if @output_flag

    if args.size > 1
      output_debug "#{func_name} *args: #{args}"

      puts "#{args[0]}: #{args[1]}" if args[1].is_a?(String)
    end

    split_file # Split file with split_format

    delete_redundant_file # Delete redundant old file
  end

  # Set log path
  def set_path(str)
    @log_path = str
    @log_folder = File.dirname(str)
    @log_file_name = File.basename(str)
  end

  # Get log path
  def get_path
    return @log_path
  end

  # Get log folder
  def get_folder
    return @log_folder
  end

  # Get log file name
  def get_file_name
    return @log_file_name
  end

  # Set format for split character
  def set_split_format(str)
    @log_split_format = str.to_s if str.is_a?(String) && !is_blank?(str)
  end

  # Get format for split character
  def get_split_format
    return @log_split_format
  end

  # Set max num of log file
  def set_log_max_num(num)
    @log_max_num = num.to_i if num.is_a?(Numeric) && !is_blank?(num)
  end

  # Get max num of log file
  def get_log_max_num
    return @log_max_num || 0
  end

end

